@@ -1,10 +1,26 @@
 from __future__ import annotations
 
+from collections import defaultdict
 from functools import update_wrapper
-from json import JSONDecoder, JSONEncoder
-from typing import Any, Callable, Iterable, List, Optional, Type, TYPE_CHECKING, Union
+from typing import Any, Callable, Iterable, List, Optional, Tuple, Type, TYPE_CHECKING, Union
 
-from .static import PackageStatic
+from .scaffold import _endpoint_from_view_func, Scaffold
+from .typing import (
+    AfterRequestCallable,
+    AfterServingCallable,
+    AfterWebsocketCallable,
+    BeforeRequestCallable,
+    BeforeServingCallable,
+    BeforeWebsocketCallable,
+    ErrorHandlerCallable,
+    TeardownCallable,
+    TemplateContextProcessorCallable,
+    TemplateFilterCallable,
+    TemplateGlobalCallable,
+    TemplateTestCallable,
+    URLDefaultCallable,
+    URLValuePreprocessorCallable,
+)
 
 if TYPE_CHECKING:
     from .app import Quart  # noqa
@@ -12,7 +28,7 @@ if TYPE_CHECKING:
 DeferredSetupFunction = Callable[["BlueprintSetupState"], Callable]
 
 
-class Blueprint(PackageStatic):
+class Blueprint(Scaffold):
     """A blueprint is a collection of application properties.
 
     The application properties include routes, error handlers, and
@@ -22,18 +38,12 @@ class Blueprint(PackageStatic):
     app.
 
     Attributes:
-        json_decoder: The decoder to use for routes in this blueprint,
-            the default, None, indicates that the app encoder should be
-            used.
-        json_encoder: The encoder to use for routes in this blueprint,
-            the default, None, indicates that the app encoder should be
-            used.
         url_prefix: An additional prefix to every route rule in the
             blueprint.
     """
 
-    json_decoder: Optional[Type[JSONDecoder]] = None
-    json_encoder: Optional[Type[JSONEncoder]] = None
+    warn_on_modifications = False
+    _got_registered_once = False
 
     def __init__(
         self,
@@ -46,9 +56,9 @@ class Blueprint(PackageStatic):
         subdomain: Optional[str] = None,
         url_defaults: Optional[dict] = None,
         root_path: Optional[str] = None,
-        **kwargs: Any,
+        cli_group: Optional[str] = Ellipsis,  # type: ignore
     ) -> None:
-        super().__init__(import_name, template_folder, root_path, static_folder, static_url_path)
+        super().__init__(import_name, static_folder, static_url_path, template_folder, root_path)
         self.name = name
         self.url_prefix = url_prefix
         self.deferred_functions: List[DeferredSetupFunction] = []
@@ -56,47 +66,11 @@ class Blueprint(PackageStatic):
         if url_defaults is None:
             url_defaults = {}
         self.url_values_defaults = url_defaults
+        self.cli_group = cli_group
+        self._blueprints: List[Tuple["Blueprint", dict]] = []
 
-    def route(
-        self,
-        rule: str,
-        methods: Optional[List[str]] = None,
-        endpoint: Optional[str] = None,
-        defaults: Optional[dict] = None,
-        host: Optional[str] = None,
-        subdomain: Optional[str] = None,
-        *,
-        provide_automatic_options: Optional[bool] = None,
-        strict_slashes: Optional[bool] = None,
-    ) -> Callable:
-        """Add a route to the blueprint.
-
-        This is designed to be used as a decorator, and has the same arguments
-        as :meth:`~quart.Quart.route`. An example usage,
-
-        .. code-block:: python
-
-            blueprint = Blueprint(__name__)
-            @blueprint.route('/')
-            def route():
-                ...
-        """
-
-        def decorator(func: Callable) -> Callable:
-            self.add_url_rule(
-                rule,
-                endpoint,
-                func,
-                methods=methods,
-                defaults=defaults,
-                host=host,
-                subdomain=subdomain,
-                provide_automatic_options=provide_automatic_options,
-                strict_slashes=strict_slashes,
-            )
-            return func
-
-        return decorator
+    def _is_setup_finished(self) -> bool:
+        return self.warn_on_modifications and self._got_registered_once
 
     def add_url_rule(
         self,
@@ -126,7 +100,7 @@ class Blueprint(PackageStatic):
             blueprint = Blueprint(__name__)
             blueprint.add_url_rule('/', route)
         """
-        endpoint = endpoint or view_func.__name__
+        endpoint = endpoint or _endpoint_from_view_func(view_func)
         if "." in endpoint:
             raise ValueError("Blueprint endpoints should not contain periods")
         self.record(
@@ -144,101 +118,6 @@ class Blueprint(PackageStatic):
             )
         )
 
-    def websocket(
-        self,
-        rule: str,
-        endpoint: Optional[str] = None,
-        defaults: Optional[dict] = None,
-        host: Optional[str] = None,
-        subdomain: Optional[str] = None,
-        *,
-        strict_slashes: Optional[bool] = None,
-    ) -> Callable:
-        """Add a websocket to the blueprint.
-
-        This is designed to be used as a decorator, and has the same arguments
-        as :meth:`~quart.Quart.websocket`. An example usage,
-
-        .. code-block:: python
-
-            blueprint = Blueprint(__name__)
-            @blueprint.websocket('/')
-            async def route():
-                ...
-        """
-
-        def decorator(func: Callable) -> Callable:
-            self.add_websocket(
-                rule,
-                endpoint,
-                func,
-                defaults=defaults,
-                host=host,
-                subdomain=subdomain,
-                strict_slashes=strict_slashes,
-            )
-            return func
-
-        return decorator
-
-    def add_websocket(
-        self,
-        rule: str,
-        endpoint: Optional[str] = None,
-        view_func: Optional[Callable] = None,
-        defaults: Optional[dict] = None,
-        host: Optional[str] = None,
-        subdomain: Optional[str] = None,
-        *,
-        strict_slashes: Optional[bool] = None,
-    ) -> None:
-        """Add a websocket rule to the blueprint.
-
-        This is designed to be used on the blueprint directly, and
-        has the same arguments as
-        :meth:`~quart.Quart.add_websocket`. An example usage,
-
-        .. code-block:: python
-
-            def route():
-                ...
-
-            blueprint = Blueprint(__name__)
-            blueprint.add_websocket('/', route)
-        """
-        return self.add_url_rule(
-            rule,
-            endpoint,
-            view_func,
-            methods={"GET"},
-            defaults=defaults,
-            host=host,
-            subdomain=subdomain,
-            provide_automatic_options=False,
-            is_websocket=True,
-            strict_slashes=strict_slashes,
-        )
-
-    def endpoint(self, endpoint: str) -> Callable:
-        """Add an endpoint to the blueprint.
-
-        This is designed to be used as a decorator, and has the same arguments
-        as :meth:`~quart.Quart.endpoint`. An example usage,
-
-        .. code-block:: python
-
-            blueprint = Blueprint(__name__)
-            @blueprint.endpoint('index')
-            def index():
-                ...
-        """
-
-        def decorator(func: Callable) -> Callable:
-            self.record_once(lambda state: state.register_endpoint(endpoint, func))
-            return func
-
-        return decorator
-
     def app_template_filter(self, name: Optional[str] = None) -> Callable:
         """Add an application wide template filter.
 
@@ -253,13 +132,15 @@ class Blueprint(PackageStatic):
                 ...
         """
 
-        def decorator(func: Callable) -> Callable:
+        def decorator(func: TemplateFilterCallable) -> TemplateFilterCallable:
             self.add_app_template_filter(func, name=name)
             return func
 
         return decorator
 
-    def add_app_template_filter(self, func: Callable, name: Optional[str] = None) -> None:
+    def add_app_template_filter(
+        self, func: TemplateFilterCallable, name: Optional[str] = None
+    ) -> None:
         """Add an application wide template filter.
 
         This is designed to be used on the blueprint directly, and
@@ -290,13 +171,13 @@ class Blueprint(PackageStatic):
                 ...
         """
 
-        def decorator(func: Callable) -> Callable:
+        def decorator(func: TemplateTestCallable) -> TemplateTestCallable:
             self.add_app_template_test(func, name=name)
             return func
 
         return decorator
 
-    def add_app_template_test(self, func: Callable, name: Optional[str] = None) -> None:
+    def add_app_template_test(self, func: TemplateTestCallable, name: Optional[str] = None) -> None:
         """Add an application wide template test.
 
         This is designed to be used on the blueprint directly, and
@@ -327,13 +208,15 @@ class Blueprint(PackageStatic):
                 ...
         """
 
-        def decorator(func: Callable) -> Callable:
+        def decorator(func: TemplateGlobalCallable) -> TemplateGlobalCallable:
             self.add_app_template_global(func, name=name)
             return func
 
         return decorator
 
-    def add_app_template_global(self, func: Callable, name: Optional[str] = None) -> None:
+    def add_app_template_global(
+        self, func: TemplateGlobalCallable, name: Optional[str] = None
+    ) -> None:
         """Add an application wide template global.
 
         This is designed to be used on the blueprint directly, and
@@ -350,42 +233,7 @@ class Blueprint(PackageStatic):
         """
         self.record_once(lambda state: state.register_template_global(func, name))
 
-    def before_request(self, func: Callable) -> Callable:
-        """Add a before request function to the Blueprint.
-
-        This is designed to be used as a decorator, and has the same arguments
-        as :meth:`~quart.Quart.before_request`. It applies only to requests that
-        are routed to an endpoint in this blueprint. An example usage,
-
-        .. code-block:: python
-
-            blueprint = Blueprint(__name__)
-            @blueprint.before_request
-            def before():
-                ...
-        """
-        self.record_once(lambda state: state.app.before_request(func, self.name))
-        return func
-
-    def before_websocket(self, func: Callable) -> Callable:
-        """Add a before request websocket to the Blueprint.
-
-        This is designed to be used as a decorator, and has the same arguments
-        as :meth:`~quart.Quart.before_websocket`. It applies only to requests that
-        are routed to an endpoint in this blueprint. An example usage,
-
-        .. code-block:: python
-
-            blueprint = Blueprint(__name__)
-            @blueprint.before_websocket
-            def before():
-                ...
-
-        """
-        self.record_once(lambda state: state.app.before_websocket(func, self.name))
-        return func
-
-    def before_app_request(self, func: Callable) -> Callable:
+    def before_app_request(self, func: BeforeRequestCallable) -> BeforeRequestCallable:
         """Add a before request function to the app.
 
         This is designed to be used as a decorator, and has the same arguments
@@ -402,8 +250,8 @@ class Blueprint(PackageStatic):
         self.record_once(lambda state: state.app.before_request(func))
         return func
 
-    def before_app_websocket(self, func: Callable) -> Callable:
-        """Add a before request websocket to the App.
+    def before_app_websocket(self, func: BeforeWebsocketCallable) -> BeforeWebsocketCallable:
+        """Add a before websocket to the App.
 
         This is designed to be used as a decorator, and has the same arguments
         as :meth:`~quart.Quart.before_websocket`. It applies to all requests to the
@@ -420,7 +268,24 @@ class Blueprint(PackageStatic):
         self.record_once(lambda state: state.app.before_websocket(func))
         return func
 
-    def before_app_first_request(self, func: Callable) -> Callable:
+    def before_app_serving(self, func: BeforeServingCallable) -> BeforeServingCallable:
+        """Add a before serving to the App.
+
+        This is designed to be used as a decorator, and has the same arguments
+        as :meth:`~quart.Quart.before_serving`. An example usage,
+
+        .. code-block:: python
+
+            blueprint = Blueprint(__name__)
+            @blueprint.before_app_serving
+            def before():
+                ...
+
+        """
+        self.record_once(lambda state: state.app.before_serving(func))
+        return func
+
+    def before_app_first_request(self, func: BeforeRequestCallable) -> BeforeRequestCallable:
         """Add a before request first function to the app.
 
         This is designed to be used as a decorator, and has the same
@@ -439,41 +304,7 @@ class Blueprint(PackageStatic):
         self.record_once(lambda state: state.app.before_first_request(func))
         return func
 
-    def after_request(self, func: Callable) -> Callable:
-        """Add an after request function to the Blueprint.
-
-        This is designed to be used as a decorator, and has the same arguments
-        as :meth:`~quart.Quart.after_request`. It applies only to requests that
-        are routed to an endpoint in this blueprint. An example usage,
-
-        .. code-block:: python
-
-            blueprint = Blueprint(__name__)
-            @blueprint.after_request
-            def after():
-                ...
-        """
-        self.record_once(lambda state: state.app.after_request(func, self.name))
-        return func
-
-    def after_websocket(self, func: Callable) -> Callable:
-        """Add an after websocket function to the Blueprint.
-
-        This is designed to be used as a decorator, and has the same arguments
-        as :meth:`~quart.Quart.after_websocket`. It applies only to requests that
-        are routed to an endpoint in this blueprint. An example usage,
-
-        .. code-block:: python
-
-            blueprint = Blueprint(__name__)
-            @blueprint.after_websocket
-            def after():
-                ...
-        """
-        self.record_once(lambda state: state.app.after_websocket(func, self.name))
-        return func
-
-    def after_app_request(self, func: Callable) -> Callable:
+    def after_app_request(self, func: AfterRequestCallable) -> AfterRequestCallable:
         """Add a after request function to the app.
 
         This is designed to be used as a decorator, and has the same arguments
@@ -490,7 +321,7 @@ class Blueprint(PackageStatic):
         self.record_once(lambda state: state.app.after_request(func))
         return func
 
-    def after_app_websocket(self, func: Callable) -> Callable:
+    def after_app_websocket(self, func: AfterWebsocketCallable) -> AfterWebsocketCallable:
         """Add an after websocket function to the App.
 
         This is designed to be used as a decorator, and has the same arguments
@@ -507,43 +338,23 @@ class Blueprint(PackageStatic):
         self.record_once(lambda state: state.app.after_websocket(func))
         return func
 
-    def teardown_request(self, func: Callable) -> Callable:
-        """Add a teardown request function to the Blueprint.
+    def after_app_serving(self, func: AfterServingCallable) -> AfterServingCallable:
+        """Add an after serving function to the App.
 
         This is designed to be used as a decorator, and has the same arguments
-        as :meth:`~quart.Quart.teardown_request`. It applies only to requests that
-        are routed to an endpoint in this blueprint. An example usage,
+        as :meth:`~quart.Quart.after_serving`. An example usage,
 
         .. code-block:: python
 
             blueprint = Blueprint(__name__)
-            @blueprint.teardown_request
-            def teardown():
+            @blueprint.after_app_serving
+            def after():
                 ...
         """
-        self.record_once(lambda state: state.app.teardown_request(func, self.name))
+        self.record_once(lambda state: state.app.after_websocket(func))
         return func
 
-    def teardown_websocket(self, func: Callable) -> Callable:
-        """Add a teardown websocket function to the Blueprint.
-
-        This is designed to be used as a decorator, and has the same
-        arguments as :meth:`~quart.Quart.teardown_websocket`. It
-        applies only to requests that are routed to an endpoint in
-        this blueprint. An example usage,
-
-        .. code-block:: python
-
-            blueprint = Blueprint(__name__)
-            @blueprint.teardown_websocket
-            def teardown():
-                ...
-
-        """
-        self.record_once(lambda state: state.app.teardown_websocket(func, self.name))
-        return func
-
-    def teardown_app_request(self, func: Callable) -> Callable:
+    def teardown_app_request(self, func: TeardownCallable) -> TeardownCallable:
         """Add a teardown request function to the app.
 
         This is designed to be used as a decorator, and has the same
@@ -561,28 +372,23 @@ class Blueprint(PackageStatic):
         self.record_once(lambda state: state.app.teardown_request(func))
         return func
 
-    def errorhandler(self, error: Union[Type[Exception], int]) -> Callable:
-        """Add an error handler function to the Blueprint.
+    def teardown_app_websocket(self, func: TeardownCallable) -> TeardownCallable:
+        """Add a teardown websocket function to the app.
 
         This is designed to be used as a decorator, and has the same
-        arguments as :meth:`~quart.Quart.errorhandler`. It applies
-        only to errors that originate in routes in this blueprint. An
+        arguments as :meth:`~quart.Quart.teardown_websocket`. It applies
+        to all requests to the app this blueprint is registered on. An
         example usage,
 
         .. code-block:: python
 
             blueprint = Blueprint(__name__)
-            @blueprint.errorhandler(404)
-            def not_found():
+            @blueprint.teardown_app_websocket
+            def teardown():
                 ...
-
         """
-
-        def decorator(func: Callable) -> Callable:
-            self.register_error_handler(error, func)
-            return func
-
-        return decorator
+        self.record_once(lambda state: state.app.teardown_websocket(func))
+        return func
 
     def app_errorhandler(self, error: Union[Type[Exception], int]) -> Callable:
         """Add an error handler function to the App.
@@ -599,48 +405,15 @@ class Blueprint(PackageStatic):
                 ...
         """
 
-        def decorator(func: Callable) -> Callable:
+        def decorator(func: ErrorHandlerCallable) -> ErrorHandlerCallable:
             self.record_once(lambda state: state.app.register_error_handler(error, func))
             return func
 
         return decorator
 
-    def register_error_handler(self, error: Union[Type[Exception], int], func: Callable) -> None:
-        """Add an error handler function to the blueprint.
-
-        This is designed to be used on the blueprint directly, and
-        has the same arguments as
-        :meth:`~quart.Quart.register_error_handler`. An example usage,
-
-        .. code-block:: python
-
-            def not_found():
-                ...
-
-            blueprint = Blueprint(__name__)
-            blueprint.register_error_handler(404, not_found)
-        """
-        self.record_once(lambda state: state.app.register_error_handler(error, func, self.name))
-
-    def context_processor(self, func: Callable) -> Callable:
-        """Add a context processor function to this blueprint.
-
-        This is designed to be used as a decorator, and has the same
-        arguments as :meth:`~quart.Quart.context_processor`. This will
-        add context to all templates rendered in this blueprint's
-        routes. An example usage,
-
-        .. code-block:: python
-
-            blueprint = Blueprint(__name__)
-            @blueprint.context_processor
-            def processor():
-                ...
-        """
-        self.record_once(lambda state: state.app.context_processor(func, self.name))
-        return func
-
-    def app_context_processor(self, func: Callable) -> Callable:
+    def app_context_processor(
+        self, func: TemplateContextProcessorCallable
+    ) -> TemplateContextProcessorCallable:
         """Add a context processor function to the app.
 
         This is designed to be used as a decorator, and has the same
@@ -657,25 +430,9 @@ class Blueprint(PackageStatic):
         self.record_once(lambda state: state.app.context_processor(func))
         return func
 
-    def url_value_preprocessor(self, func: Callable) -> Callable:
-        """Add a url value preprocessor.
-
-        This is designed to be used as a decorator, and has the same
-        arguments as :meth:`~quart.Quart.url_value_preprocessor`. This
-        will apply to urls in this blueprint. An example usage,
-
-        .. code-block:: python
-
-            blueprint = Blueprint(__name__)
-            @blueprint.url_value_preprocessor
-            def processor(endpoint, view_args):
-                ...
-
-        """
-        self.record_once(lambda state: state.app.url_value_preprocessor(func, self.name))
-        return func
-
-    def app_url_value_preprocessor(self, func: Callable) -> Callable:
+    def app_url_value_preprocessor(
+        self, func: URLValuePreprocessorCallable
+    ) -> URLValuePreprocessorCallable:
         """Add a url value preprocessor.
 
         This is designed to be used as a decorator, and has the same
@@ -694,25 +451,7 @@ class Blueprint(PackageStatic):
         self.record_once(lambda state: state.app.url_value_preprocessor(func))
         return func
 
-    def url_defaults(self, func: Callable) -> Callable:
-        """Add a url default preprocessor.
-
-        This is designed to be used as a decorator, and has the same
-        arguments as :meth:`~quart.Quart.url_defaults`. This will
-        apply to urls in this blueprint. An example usage,
-
-        .. code-block:: python
-
-            blueprint = Blueprint(__name__)
-            @blueprint.url_defaults
-            def default(endpoint, values):
-                ...
-
-        """
-        self.record_once(lambda state: state.app.url_defaults(func, self.name))
-        return func
-
-    def app_url_defaults(self, func: Callable) -> Callable:
+    def app_url_defaults(self, func: URLDefaultCallable) -> URLDefaultCallable:
         """Add a url default preprocessor.
 
         This is designed to be used as a decorator, and has the same
@@ -741,9 +480,17 @@ class Blueprint(PackageStatic):
             if state.first_registration:
                 func(state)
 
-        self.record(update_wrapper(wrapper, func))
+        return self.record(update_wrapper(wrapper, func))
 
-    def register(self, app: "Quart", options: dict, first_registration: bool = False) -> None:
+    def register_blueprint(self, blueprint: "Blueprint", **options: Any) -> None:
+        """Register a :class:`~quart.Blueprint` on this blueprint.
+
+        Keyword arguments passed to this method will override the
+        defaults set on the blueprint.
+        """
+        self._blueprints.append((blueprint, options))
+
+    def register(self, app: "Quart", options: dict) -> None:
         """Register this blueprint on the app given.
 
         Arguments:
@@ -753,6 +500,19 @@ class Blueprint(PackageStatic):
             first_registration: Whether this is the first time this
                 blueprint has been registered on the application.
         """
+        first_registration = False
+
+        if self.name in app.blueprints and app.blueprints[self.name] is not self:
+            raise RuntimeError(
+                f"Blueprint name '{self.name}' "
+                f"is already registered by {app.blueprints[self.name]}. "
+                "Blueprints must have unique names"
+            )
+        else:
+            app.blueprints[self.name] = self
+            first_registration = True
+
+        self._got_registered_once = True
         state = self.make_setup_state(app, options, first_registration)
 
         if self.has_static_folder:
@@ -762,9 +522,67 @@ class Blueprint(PackageStatic):
                 endpoint="static",
             )
 
+        if first_registration:
+
+            def extend(bp_dict: dict, parent_dict: dict) -> None:
+                for key, values in bp_dict.items():
+                    key = self.name if key is None else f"{self.name}.{key}"
+                    parent_dict[key].extend(values)
+
+            for key, value in self.error_handler_spec.items():
+                key = self.name if key is None else f"{self.name}.{key}"
+                value = defaultdict(
+                    dict,
+                    {
+                        code: {exc_class: func for exc_class, func in code_values.items()}
+                        for code, code_values in value.items()
+                    },
+                )
+                app.error_handler_spec[key] = value
+
+            for endpoint, func in self.view_functions.items():
+                app.view_functions[endpoint] = func
+
+            extend(self.before_request_funcs, app.before_request_funcs)
+            extend(self.before_websocket_funcs, app.before_websocket_funcs)
+            extend(self.after_request_funcs, app.after_request_funcs)
+            extend(self.after_websocket_funcs, app.after_websocket_funcs)
+            extend(
+                self.teardown_request_funcs,
+                app.teardown_request_funcs,
+            )
+            extend(
+                self.teardown_websocket_funcs,
+                app.teardown_websocket_funcs,
+            )
+            extend(self.url_default_functions, app.url_default_functions)
+            extend(self.url_value_preprocessors, app.url_value_preprocessors)
+            extend(self.template_context_processors, app.template_context_processors)
+
         for func in self.deferred_functions:
             func(state)
 
+        cli_resolved_group = options.get("cli_group", self.cli_group)
+
+        if self.cli.commands:
+            if cli_resolved_group is None:
+                app.cli.commands.update(self.cli.commands)
+            elif cli_resolved_group is Ellipsis:
+                self.cli.name = self.name
+                app.cli.add_command(self.cli)
+            else:
+                self.cli.name = cli_resolved_group
+                app.cli.add_command(self.cli)
+
+        for blueprint, bp_options in self._blueprints:
+            url_prefix = options.get("url_prefix", "")
+            if "url_prefix" in bp_options:
+                url_prefix = url_prefix.rstrip("/") + "/" + bp_options["url_prefix"].lstrip("/")
+
+            bp_options["url_prefix"] = url_prefix
+            bp_options["name_prefix"] = options.get("name_prefix", "") + self.name + "."
+            blueprint.register(app, bp_options)
+
     def make_setup_state(
         self, app: "Quart", options: dict, first_registration: bool = False
     ) -> "BlueprintSetupState":
@@ -804,6 +622,7 @@ class BlueprintSetupState:
         self.subdomain = options.get("subdomain") or blueprint.subdomain
         self.url_defaults = dict(self.blueprint.url_values_defaults)
         self.url_defaults.update(options.get("url_defaults", {}) or {})
+        self.name_prefix = self.options.get("name_prefix", "")
 
     def add_url_rule(
         self,
@@ -820,10 +639,10 @@ class BlueprintSetupState:
         strict_slashes: Optional[bool] = None,
     ) -> None:
         if self.url_prefix is not None:
-            path = f"{self.url_prefix}{path}"
+            path = f"{self.url_prefix.rstrip('/')}/{path.lstrip('/')}"
         if subdomain is None:
             subdomain = self.subdomain
-        endpoint = f"{self.blueprint.name}.{endpoint}"
+        endpoint = f"{self.name_prefix}{self.blueprint.name}.{endpoint}"
         url_defaults = self.url_defaults
         if defaults is not None:
             url_defaults = {**url_defaults, **defaults}
@@ -840,10 +659,7 @@ class BlueprintSetupState:
             strict_slashes=strict_slashes,
         )
 
-    def register_endpoint(self, endpoint: str, func: Callable) -> None:
-        self.app.view_functions[endpoint] = func
-
-    def register_template_filter(self, func: Callable, name: Optional[str]) -> None:
+    def register_template_filter(self, func: TemplateFilterCallable, name: Optional[str]) -> None:
         self.app.add_template_filter(func, name)
 
     def register_template_test(self, func: Callable, name: Optional[str]) -> None:
@@ -851,3 +667,15 @@ class BlueprintSetupState:
 
     def register_template_global(self, func: Callable, name: Optional[str]) -> None:
         self.app.add_template_global(func, name)
+
+
+def _merge_dict_of_lists(name: str, self_dict: dict, app_dict: dict) -> None:
+    for key, values in self_dict.items():
+        key = name if key is None else f"{name}.{key}"
+        app_dict[key].extend(values)
+
+
+def _merge_dict_of_dicts(name: str, self_dict: dict, app_dict: dict) -> None:
+    for key, value in self_dict.items():
+        key = name if key is None else f"{name}.{key}"
+        app_dict[key] = value
