@@ -4,17 +4,12 @@ from functools import wraps
 from types import TracebackType
 from typing import Any, Callable, cast, Iterator, List, Optional, TYPE_CHECKING  # noqa: F401
 
-from werkzeug.exceptions import (
-    BadRequest as WBadRequest,
-    MethodNotAllowed as WMethodNotAllowed,
-    NotFound as WNotFound,
-)
-from werkzeug.routing import RequestRedirect as WRequestRedirect
+from werkzeug.exceptions import HTTPException
 
-from .exceptions import BadRequest, MethodNotAllowed, NotFound, RedirectRequired
 from .globals import _app_ctx_stack, _request_ctx_stack, _websocket_ctx_stack
-from .sessions import Session  # noqa
+from .sessions import SessionMixin  # noqa
 from .signals import appcontext_popped, appcontext_pushed
+from .typing import AfterRequestCallable, AfterWebsocketCallable
 from .wrappers import BaseRequestWebsocket, Request, Websocket
 
 if TYPE_CHECKING:
@@ -35,9 +30,7 @@ class _BaseRequestWebsocketContext:
         self,
         app: "Quart",
         request_websocket: BaseRequestWebsocket,
-        session: Optional[Session] = None,
-        *,
-        _preserve: bool = False,
+        session: Optional[SessionMixin] = None,
     ) -> None:
         self.app = app
         self.request_websocket = request_websocket
@@ -45,10 +38,17 @@ class _BaseRequestWebsocketContext:
         self.request_websocket.routing_exception = None
         self.session = session
         self.preserved = False
-        self._should_preserve = _preserve
 
         self.match_request()
 
+    @property
+    def g(self) -> AppContext:
+        return _app_ctx_stack.top.g
+
+    @g.setter
+    def g(self, value: AppContext) -> None:
+        _app_ctx_stack.top.g = value
+
     def copy(self) -> "_BaseRequestWebsocketContext":
         return self.__class__(self.app, self.request_websocket, self.session)
 
@@ -63,38 +63,40 @@ class _BaseRequestWebsocketContext:
             (
                 self.request_websocket.url_rule,
                 self.request_websocket.view_args,
-            ) = self.url_adapter.match(
+            ) = self.url_adapter.match(  # type: ignore
                 return_rule=True
             )  # noqa
-        except WBadRequest:
-            self.request_websocket.routing_exception = BadRequest()
-        except WNotFound:
-            self.request_websocket.routing_exception = NotFound()
-        except WMethodNotAllowed as error:
-            new_error = MethodNotAllowed(error.valid_methods)
-            self.request_websocket.routing_exception = new_error
-        except WRequestRedirect as error:
-            new_error = RedirectRequired(error.new_url)  # type: ignore
-            self.request_websocket.routing_exception = new_error
+        except HTTPException as exception:
+            self.request_websocket.routing_exception = exception
 
-    async def pop(self, exc: BaseException) -> None:
-        await _app_ctx_stack.top.pop(exc)
-
-    async def auto_pop(self, exc: BaseException) -> None:
-        if self._should_preserve:
-            self.preserved = True
-        else:
-            await self.pop(exc)
-
-    async def __aenter__(self) -> "_BaseRequestWebsocketContext":
+    async def push(self) -> None:
         app_ctx = _app_ctx_stack.top
         if app_ctx is None:
             app_ctx = self.app.app_context()
         await app_ctx.push()
 
-        self.session = await self.app.open_session(self.request_websocket)
         if self.session is None:
-            self.session = await self.app.make_null_session()
+            session_interface = self.app.session_interface
+            self.session = await self.app.ensure_async(session_interface.open_session)(
+                self.app, self.request_websocket
+            )
+
+            if self.session is None:
+                self.session = await session_interface.make_null_session(self.app)
+
+    async def pop(self, exc: BaseException) -> None:
+        await _app_ctx_stack.top.pop(exc)
+
+    async def auto_pop(self, exc: Optional[BaseException]) -> None:
+        if self.request_websocket.scope.get("_quart._preserve_context", False) or (  # type: ignore
+            exc is not None and self.app.preserve_context_on_exception
+        ):
+            self.preserved = True
+        else:
+            await self.pop(exc)
+
+    async def __aenter__(self) -> "_BaseRequestWebsocketContext":
+        await self.push()
         return self
 
     async def __aexit__(self, exc_type: type, exc_value: BaseException, tb: TracebackType) -> None:
@@ -117,25 +119,27 @@ class RequestContext(_BaseRequestWebsocketContext):
         self,
         app: "Quart",
         request: Request,
-        session: Optional[Session] = None,
-        *,
-        _preserve: bool = False,
+        session: Optional[SessionMixin] = None,
     ) -> None:
-        super().__init__(app, request, session, _preserve=_preserve)
-        self._after_request_functions: List[Callable] = []
+        super().__init__(app, request, session)
+        self.flashes = None
+        self._after_request_functions: List[AfterRequestCallable] = []
 
     @property
     def request(self) -> Request:
         return cast(Request, self.request_websocket)
 
+    async def push(self) -> None:
+        await super().push()
+        _request_ctx_stack.push(self)
+
     async def pop(self, exc: BaseException) -> None:
         await self.app.do_teardown_request(exc, self)
         _request_ctx_stack.pop()
         await super().pop(exc)
 
     async def __aenter__(self) -> "RequestContext":
-        await super().__aenter__()
-        _request_ctx_stack.push(self)
+        await self.push()
         return self
 
 
@@ -155,25 +159,26 @@ class WebsocketContext(_BaseRequestWebsocketContext):
         self,
         app: "Quart",
         request: Websocket,
-        session: Optional[Session] = None,
-        *,
-        _preserve: bool = False,
+        session: Optional[SessionMixin] = None,
     ) -> None:
-        super().__init__(app, request, session, _preserve=_preserve)
-        self._after_websocket_functions: List[Callable] = []
+        super().__init__(app, request, session)
+        self._after_websocket_functions: List[AfterWebsocketCallable] = []
 
     @property
     def websocket(self) -> Websocket:
         return cast(Websocket, self.request_websocket)
 
+    async def push(self) -> None:
+        await super().push()
+        _websocket_ctx_stack.push(self)
+
     async def pop(self, exc: BaseException) -> None:
         await self.app.do_teardown_websocket(exc, self)
         _websocket_ctx_stack.pop()
         await super().pop(exc)
 
     async def __aenter__(self) -> "WebsocketContext":
-        await super().__aenter__()
-        _websocket_ctx_stack.push(self)
+        await self.push()
         return self
 
 
@@ -224,7 +229,7 @@ class AppContext:
         await self.pop(exc_value)
 
 
-def after_this_request(func: Callable) -> Callable:
+def after_this_request(func: AfterRequestCallable) -> AfterRequestCallable:
     """Schedule the func to be called after the current request.
 
     This is useful in situations whereby you want an after request
@@ -244,7 +249,7 @@ def after_this_request(func: Callable) -> Callable:
     return func
 
 
-def after_this_websocket(func: Callable) -> Callable:
+def after_this_websocket(func: AfterWebsocketCallable) -> AfterWebsocketCallable:
     """Schedule the func to be called after the current websocket.
 
     This is useful in situations whereby you want an after websocket
