@@ -1,16 +1,21 @@
 from __future__ import annotations
 
 import asyncio
-import io
-from cgi import FieldStorage, parse_header
-from typing import Any, AnyStr, Awaitable, Callable, Generator, List, Optional, Union
-from urllib.parse import parse_qs
+from typing import Any, AnyStr, Awaitable, Callable, Dict, Generator, NoReturn, Optional, overload
 
+from hypercorn.typing import HTTPScope
 from werkzeug.datastructures import CombinedMultiDict, Headers, MultiDict
+from werkzeug.exceptions import BadRequest, RequestEntityTooLarge, RequestTimeout
 
-from .base import BaseRequestWebsocket, JSONMixin
-from ..datastructures import FileStorage
-from ..json import dumps, loads
+from .base import BaseRequestWebsocket
+from .. import json
+from ..formparser import FormDataParser
+from ..globals import current_app
+
+try:
+    from typing import Literal
+except ImportError:
+    from typing_extensions import Literal  # type: ignore
 
 SERVER_PUSH_HEADERS_TO_COPY = {
     "accept",
@@ -54,8 +59,6 @@ class Body:
             and max_content_length is not None
             and expected_content_length > max_content_length
         ):
-            from ..exceptions import RequestEntityTooLarge  # noqa Avoiding circular import
-
             self._must_raise = RequestEntityTooLarge()
 
     def __aiter__(self) -> "Body":
@@ -83,7 +86,6 @@ class Body:
         # Must check the _must_raise before and after waiting on the
         # completion event as it may change whilst waiting and the
         # event may not be set if there is already an issue.
-
         if self._must_raise is not None:
             raise self._must_raise
 
@@ -99,8 +101,6 @@ class Body:
         self._data.extend(data)
         self._has_data.set()
         if self._max_content_length is not None and len(self._data) > self._max_content_length:
-            from ..exceptions import RequestEntityTooLarge  # noqa Avoiding circular import
-
             self._must_raise = RequestEntityTooLarge()
             self.set_complete()
 
@@ -113,8 +113,11 @@ class Body:
         self.append(data)
         self.set_complete()
 
+    def clear(self) -> None:
+        self._data.clear()
 
-class Request(BaseRequestWebsocket, JSONMixin):
+
+class Request(BaseRequestWebsocket):
     """This class represents a request.
 
     It can be subclassed and the subclassed used in preference by
@@ -123,9 +126,15 @@ class Request(BaseRequestWebsocket, JSONMixin):
 
     Attributes:
         body_class: The class to store the body data within.
+        form_data_parser_class: Can be overridden to implement a
+            different form data parsing.
+        json_module: A custom json decoding/encoding module, it should
+            have `dump`, `dumps`, `load`, and `loads` methods
     """
 
     body_class = Body
+    form_data_parser_class = FormDataParser
+    json_module = json
 
     def __init__(
         self,
@@ -136,11 +145,11 @@ class Request(BaseRequestWebsocket, JSONMixin):
         headers: Headers,
         root_path: str,
         http_version: str,
+        scope: HTTPScope,
         *,
         max_content_length: Optional[int] = None,
         body_timeout: Optional[int] = None,
         send_push_promise: Callable[[str, Headers], Awaitable[None]],
-        scope: Optional[dict] = None,
     ) -> None:
         """Create a request object.
 
@@ -163,16 +172,55 @@ class Request(BaseRequestWebsocket, JSONMixin):
                 off of this request (HTTP/2 feature).
             scope: Underlying ASGI scope dictionary.
         """
-        super().__init__(method, scheme, path, query_string, headers, root_path, http_version)
+        super().__init__(
+            method, scheme, path, query_string, headers, root_path, http_version, scope
+        )
         self.body_timeout = body_timeout
         self.body = self.body_class(self.content_length, max_content_length)
+        self._cached_json: Dict[bool, Any] = {False: Ellipsis, True: Ellipsis}
         self._form: Optional[MultiDict] = None
         self._files: Optional[MultiDict] = None
         self._send_push_promise = send_push_promise
-        self.scope = scope
 
-    async def get_data(self, raw: bool = True) -> AnyStr:
-        """The request body data."""
+    @property
+    async def stream(self) -> NoReturn:
+        raise NotImplementedError("Use body instead")
+
+    @property
+    async def data(self) -> bytes:
+        return await self.get_data(as_text=False, parse_form_data=True)
+
+    @overload
+    async def get_data(self, cache: bool, as_text: Literal[False], parse_form_data: bool) -> bytes:
+        ...
+
+    @overload
+    async def get_data(self, cache: bool, as_text: Literal[True], parse_form_data: bool) -> str:
+        ...
+
+    @overload
+    async def get_data(
+        self, cache: bool = True, as_text: bool = False, parse_form_data: bool = False
+    ) -> AnyStr:
+        ...
+
+    async def get_data(
+        self, cache: bool = True, as_text: bool = False, parse_form_data: bool = False
+    ) -> AnyStr:
+        """Get the request body data.
+
+        Arguments:
+            cache: If False the body data will be cleared, resulting in any
+                subsequent calls returning an empty AnyStr and reducing
+                memory usage.
+            as_text: If True the data is returned as a decoded string,
+                otherwise raw bytes are returned.
+            parse_form_data: Parse the data as form data first, return any
+                remaining data.
+        """
+        if parse_form_data:
+            await self._load_form_data()
+
         try:
             body_future = asyncio.ensure_future(self.body)
             raw_data = await asyncio.wait_for(body_future, timeout=self.body_timeout)
@@ -183,23 +231,34 @@ class Request(BaseRequestWebsocket, JSONMixin):
             except asyncio.CancelledError:
                 pass
 
-            from ..exceptions import RequestTimeout  # noqa Avoiding circular import
-
             raise RequestTimeout()
-
-        if raw:
-            return raw_data
         else:
-            return raw_data.decode(self.charset)
+            if not cache:
+                self.body.clear()
 
-    @property
-    async def data(self) -> bytes:
-        return await self.get_data()
+            if as_text:
+                return raw_data.decode(self.charset, self.encoding_errors)
+            else:
+                return raw_data
 
     @property
     async def values(self) -> CombinedMultiDict:
-        form = await self.form
-        return CombinedMultiDict([self.args, form])
+        sources = [self.args]
+        if self.method != "GET":
+            # Whilst GET requests are allowed to have a body, most
+            # implementations do not allow this hence this
+            # inconsistency may result in confusing values.
+            form = await self.form
+            sources.append(form)
+
+        multidict_sources = []
+        for source in sources:
+            if not isinstance(source, MultiDict):
+                multidict_sources.append(MultiDict(source))
+            else:
+                multidict_sources.append(source)
+
+        return CombinedMultiDict(multidict_sources)
 
     @property
     async def form(self) -> MultiDict:
@@ -207,8 +266,7 @@ class Request(BaseRequestWebsocket, JSONMixin):
 
         Note file data is present in the :attr:`files`.
         """
-        if self._form is None:
-            await self._load_form_data()
+        await self._load_form_data()
         return self._form
 
     @property
@@ -219,80 +277,77 @@ class Request(BaseRequestWebsocket, JSONMixin):
         mimetype was ``enctype="multipart/form-data"`` and the method
         POST, PUT, or PATCH.
         """
-        if self._files is None:
-            await self._load_form_data()
+        await self._load_form_data()
         return self._files
 
+    def make_form_data_parser(self) -> FormDataParser:
+        return self.form_data_parser_class(
+            charset=self.charset,
+            errors=self.encoding_errors,
+            max_content_length=self.max_content_length,
+            cls=self.parameter_storage_class,
+        )
+
     async def _load_form_data(self) -> None:
-        raw_data: bytes = await self.get_data(raw=True)
-        self._form = MultiDict()
-        self._files = MultiDict()
-        content_header = self.content_type
-        if content_header is None:
-            return
-        content_type, parameters = parse_header(content_header)
-        if content_type == "application/x-www-form-urlencoded":
-            try:
-                data = raw_data.decode(parameters.get("charset", "utf-8"))
-            except UnicodeDecodeError:
-                from ..exceptions import BadRequest  # noqa Avoiding circular import
-
-                raise BadRequest()
-            for key, values in parse_qs(data, keep_blank_values=True).items():
-                for value in values:
-                    self._form.add(key, value)
-        elif content_type == "multipart/form-data":
-            field_storage = FieldStorage(
-                io.BytesIO(raw_data),
-                headers={name.lower(): value for name, value in self.headers.items()},
-                environ={"REQUEST_METHOD": "POST"},
-                limit=len(raw_data),
-            )
-            for key in field_storage:  # type: ignore
-                field_storage_key = field_storage[key]
-                if isinstance(field_storage_key, list):
-                    for item in field_storage_key:
-                        self._load_field_storage(key, item)
-                else:
-                    self._load_field_storage(key, field_storage_key)
-
-    def _load_field_storage(self, key: str, field_storage: FieldStorage) -> None:
-        if isinstance(field_storage, FieldStorage) and field_storage.filename is not None:
-            self._files.add(
-                key,
-                FileStorage(
-                    io.BytesIO(field_storage.file.read()),
-                    field_storage.filename,
-                    field_storage.name,  # type: ignore
-                    field_storage.type,
-                    field_storage.headers,  # type: ignore
-                ),
-            )
-        else:
-            self._form.add(key, field_storage.value)
+        parser = self.make_form_data_parser()
+        self._form, self._files = await asyncio.wait_for(
+            parser.parse(
+                self.body,
+                self.mimetype,
+                self.content_length,
+                self.mimetype_params,
+            ),
+            timeout=self.body_timeout,
+        )
 
     @property
-    def content_encoding(self) -> Optional[str]:
-        return self.headers.get("Content-Encoding")
+    async def json(self) -> Any:
+        return await self.get_json()
 
-    @property
-    def content_length(self) -> Optional[int]:
-        if "Content-Length" in self.headers:
-            return int(self.headers["Content-Length"])
-        else:
+    async def get_json(self, force: bool = False, silent: bool = False, cache: bool = True) -> Any:
+        """Parses the body data as JSON and returns it.
+
+        Arguments:
+            force: Force JSON parsing even if the mimetype is not JSON.
+            silent: Do not trigger error handling if parsing fails, without
+                this the :meth:`on_json_loading_failed` will be called on
+                error.
+            cache: Cache the parsed JSON on this request object.
+        """
+        if cache and self._cached_json[silent] is not Ellipsis:
+            return self._cached_json[silent]
+
+        if not (force or self.is_json):
             return None
 
-    @property
-    def content_md5(self) -> Optional[str]:
-        return self.headers.get("Content-md5")
+        data = await self.get_data(cache=cache, as_text=True)
 
-    @property
-    def content_type(self) -> Optional[str]:
-        return self.headers.get("Content-Type")
+        try:
+            result = self.json_module.loads(data)
+        except ValueError as error:
+            if silent:
+                result = None
+            else:
+                result = self.on_json_loading_failed(error)
 
-    async def _load_json_data(self) -> str:
-        """Return the data after decoding."""
-        return await self.get_data(raw=False)
+        if cache:
+            self._cached_json[silent] = result
+
+        return result
+
+    def on_json_loading_failed(self, error: Exception) -> Any:
+        """Handle a JSON parsing error.
+
+        Arguments:
+            error: The exception raised during parsing.
+        Returns:
+            Any value returned (if overidden) will be used as the
+            default for any get_json calls.
+        """
+        if current_app and current_app.debug:
+            raise BadRequest(f"Failed to decode JSON: {error}")
+
+        raise BadRequest()
 
     async def send_push_promise(self, path: str) -> None:
         headers = Headers()
@@ -300,85 +355,3 @@ class Request(BaseRequestWebsocket, JSONMixin):
             for value in self.headers.getlist(name):
                 headers.add(name, value)
         await self._send_push_promise(path, headers)
-
-    def __repr__(self) -> str:
-        return f"{self.__class__.__name__}({self.method}, {self.path})"
-
-
-class Websocket(BaseRequestWebsocket):
-    def __init__(
-        self,
-        path: str,
-        query_string: bytes,
-        scheme: str,
-        headers: Headers,
-        root_path: str,
-        http_version: str,
-        subprotocols: List[str],
-        receive: Callable,
-        send: Callable,
-        accept: Callable,
-    ) -> None:
-        """Create a request object.
-
-        Arguments:
-            path: The full unquoted path of the request.
-            query_string: The raw bytes for the query string part.
-            scheme: The scheme used for the request.
-            headers: The request headers.
-            root_path: The root path that should be prepended to all
-                routes.
-            http_version: The HTTP version of the request.
-            subprotocols: The subprotocols requested.
-            receive: Returns an awaitable of the current data
-
-            accept: Idempotent callable to accept the websocket connection.
-        """
-        super().__init__("GET", scheme, path, query_string, headers, root_path, http_version)
-        self._accept = accept
-        self._receive = receive
-        self._send = send
-        self._subprotocols = subprotocols
-
-    @property
-    def requested_subprotocols(self) -> List[str]:
-        return self._subprotocols
-
-    async def receive(self) -> AnyStr:
-        await self.accept()
-        return await self._receive()
-
-    async def send(self, data: AnyStr) -> None:
-        # Must allow for the event loop to act if the user has say
-        # setup a tight loop sending data over a websocket (as in the
-        # example). So yield via the sleep.
-        await asyncio.sleep(0)
-        await self.accept()
-        await self._send(data)
-
-    async def receive_json(self) -> Any:
-        data = await self.receive()
-        return loads(data)
-
-    async def send_json(self, data: Any) -> None:
-        raw = dumps(data)
-        await self.send(raw)
-
-    async def accept(
-        self, headers: Optional[Union[dict, Headers]] = None, subprotocol: Optional[str] = None
-    ) -> None:
-        """Manually chose to accept the websocket connection.
-
-        Arguments:
-            headers: Additional headers to send with the acceptance
-                response.
-            subprotocol: The chosen subprotocol, optional.
-        """
-        if headers is None:
-            headers_ = Headers()
-        else:
-            headers_ = Headers(headers)
-        await self._accept(headers_, subprotocol)
-
-    def __repr__(self) -> str:
-        return f"{self.__class__.__name__}({self.path})"
